from dataclasses import dataclass
from enum import IntEnum
from typing import Optional

from tplink_wr.parse.utils import extract_vars
from tplink_wr.router import RouterInterface

from .fetcher import Fetcher


class RouterType(IntEnum):
    WAN = 1
    WITH_3G = 2
    APC = 4
    PURE_3G = 16


class WLANType(IntEnum):
    UNKNOWN = 0
    B = 1
    G = 2
    N = 3
    BG_MIXED = 4
    GN_MIXED = 5
    BGN_MIXED = 6
    A = 7
    N_DUPL = 8
    AN_MIXED = 9


class WLANChannelWidth(IntEnum):
    UNKNOWN = 0
    MHZ_20 = 1
    AUTO = 2
    MHZ_40 = 3


class WDSStatus(IntEnum):
    INIT = 0
    SCAN = 1
    JOIN = 2
    AUTH = 3
    ASSOC = 4
    RUN = 5
    DISABLE = 6


class WANLinkStatus(IntEnum):
    UNKNOWN = 0
    DISABLED = 1
    TIMEOUT = 2
    LINK_DOWN = 3
    LINK_UP = 4


class WANType(IntEnum):
    UNKNOWN = 0
    DYNAMIC = 1
    STATIC = 2
    PPPOE = 3
    DYNAMIC_1X = 4
    STATIC_1X = 5
    BIGPOND = 6
    L2TP = 7
    PPTP = 8


@dataclass
class LANStatus:
    mac: str
    ip: str
    mask: str


@dataclass
class WLANStatus:
    enabled: bool
    name: str
    type: WLANType
    channel_manual: Optional[int]
    channel_auto: Optional[int]
    channel_width: WLANChannelWidth
    mac: str
    ip: str
    wds_status: WDSStatus


@dataclass
class WANStatus:
    link_status: WANLinkStatus
    mac: str
    ip: str
    type: WANType
    mask: str
    gateway: str
    dns: str


@dataclass
class GeneralStatus(Fetcher):
    wireless: bool
    uptime: int
    firmware: str
    hardware: str
    device_type: RouterType
    mode_3g: bool

    rx_bytes: int
    tx_bytes: int
    rx_packets: int
    tx_packets: int

    lan: LANStatus
    wlan: Optional[WLANStatus]
    wan: list[WANStatus]

    @classmethod
    def fetch(cls, router: RouterInterface):
        doc = router.page("StatusRpm")
        general, lan, wlan, statist, wan = extract_vars(doc, [
            "statusPara", "lanPara", "wlanPara", "statistList", "wanPara"
        ]).values()

        status = {
            **cls._parse_status(general),
            **cls._parse_statist(statist),
            "lan": cls._parse_lan(lan),
            "wlan": cls._parse_wlan(wlan, general),
            "wan": cls._parse_wan(wan, general),
        }

        status_obj = cls(**status)
        return status_obj

    @staticmethod
    def _parse_status(status) -> dict:
        return {
            "wireless": bool(status[0]),
            "uptime": status[4],
            "firmware": status[5],
            "hardware": status[6],
            "device_type": RouterType(status[7]),
            "mode_3g": bool(status[8]),
        }

    @staticmethod
    def _parse_statist(statist) -> dict:
        statist_parse = lambda value: int(value.replace(",", ""))
        return {
            "rx_bytes": statist_parse(statist[0]),
            "tx_bytes": statist_parse(statist[1]),
            "rx_packets": statist_parse(statist[2]),
            "tx_packets": statist_parse(statist[3]),
        }

    @staticmethod
    def _parse_lan(lan) -> LANStatus:
        return LANStatus(
            mac=lan[0],
            ip=lan[1],
            mask=lan[2],
        )

    @staticmethod
    def _parse_wlan(wlan, general) -> Optional[WLANStatus]:
        if not general[0]:
            return None

        return WLANStatus(
            enabled=bool(wlan[0]),
            name=wlan[1],
            type=WLANType(wlan[3]),
            channel_manual=None if wlan[2] == 15 else wlan[2],
            channel_width=WLANChannelWidth(wlan[6]),
            channel_auto=wlan[9],
            mac=wlan[4],
            ip=wlan[5],
            wds_status=WDSStatus(wlan[10]),
        )

    @staticmethod
    def _parse_wan(wan, general) -> list[WANStatus]:
        wan_count = general[1]
        wan_params_per_item = general[2]
        assert len(wan) == wan_count * wan_params_per_item
        assert wan_params_per_item >= 12

        result = []
        for i in range(wan_count):
            base = i * wan_params_per_item
            wan_status = WANStatus(
                link_status=WANLinkStatus(wan[base]),
                mac=wan[base+1],
                ip=wan[base+2],
                type=WANType(wan[base+3]),
                mask=wan[base+4],
                gateway=wan[base+7],
                dns=wan[base+11],
            )
            result.append(wan_status)

        return result



from dataclasses import dataclass

from tplink_wr.parse.utils import extract_vars
from tplink_wr.router import RouterInterface

from .fetcher import Fetcher


@dataclass
class DHCPLeases(Fetcher):
    leases: list

    @classmethod
    def fetch(cls, router: RouterInterface):
        doc = router.page("AssignedIpAddrListRpm")
        dyn_para, dyn_list = extract_vars(doc, ["DHCPDynPara", "DHCPDynList"]).values()

        leases_count = dyn_para[0]
        params_per_lease = dyn_para[1]
        assert len(dyn_list) == leases_count * params_per_lease
        assert params_per_lease >= 4

        leases_raw = []
        for i in range(leases_count):
            base = i * params_per_lease

            expire = None
            expire_raw = dyn_list[base+3]
            if ":" in expire_raw:
                hours, minutes, seconds = [
                    int(part)
                    for part in expire_raw.split(":")
                ]
                expire = seconds + minutes * 60 + hours * 60 * 60

            leases_raw.append({
                "name": dyn_list[base],
                "mac": dyn_list[base+1],
                "ip": dyn_list[base+2],
                "expire": expire,
            })

        leases = cls(leases=leases_raw)
        return leases



from . import html, js


def extract_vars(document: str, vars: list[str]):
    js_vars = {
        var: None
        for var in vars
    }

    scripts = html.find_scripts(document)
    for script in scripts:
        for var in js_vars:
            if js_vars[var] is None:
                js_vars[var] = js.get_var(var, script)

    return js_vars



from abc import ABC, abstractmethod
from typing import Optional
import base64
import hashlib
import re

import requests

from . import exceptions
from .parse.html import find_scripts


class RouterInterface(ABC):
    @abstractmethod
    def page(self, name: str, params: Optional[dict] = None) -> str:
        pass


class RouterSession(RouterInterface):
    def __init__(self, host: str, username: str, password: str, *, auto_reauth: bool = True, auth_retries: int = 3, timeout: Optional[int] = None):
        self.host = host
        self.auto_reauth = bool(auto_reauth)
        self.auth_retries = max(auth_retries, 0)
        self.timeout = timeout

        password_hash = hashlib.md5(password.encode()).hexdigest()
        basic_raw = f"{username}:{password_hash}".encode()
        basic_token = base64.b64encode(basic_raw).decode("utf-8")

        self.session = requests.Session()
        self.session.cookies["Authorization"] = f"Basic {basic_token}"
        self.refresh_token()

    def is_session_valid(self) -> bool:
        url = self.page_url("Index")
        resp = self._get(url)
        reauth = self._is_reauth_doc(resp.text)
        return not reauth

    def refresh_token(self):
        attempts = self.auth_retries+1
        for retry in range(attempts):
            resp = self._get(f"{self.base_url()}/userRpm/LoginRpm.htm?Save=Save")

            match = re.search(f"{self.host}/(\w+)/", resp.text)
            if match:
                self.token = match.group(1)
                if self.is_session_valid():
                    return

        raise exceptions.AuthError(f"Failed to get auth token with specified username and password after {attempts} attempts")

    def base_url(self) -> str:
        return f"http://{self.host}"

    def page_url(self, name: str) -> str:
        return f"{self.base_url()}/{self.token}/userRpm/{name}.htm"

    def page(self, name: str, params: Optional[dict] = None) -> str:
        retry = False
        while True:
            doc = self._page_load_attempt(name, params)
            if not self._is_reauth_doc(doc):
                break

            if retry or not self.auto_reauth:
                raise exceptions.PageLoadError(f"Failed to load page {name}. Firmware of the router may not support this feature")

            retry = True
            self.refresh_token()

        return doc

    def _page_load_attempt(self, name: str, params: Optional[dict] = None) -> str:
        url = self.page_url(name)
        referer = self.page_url("MenuRpm")

        resp = self._get(url, params=params, headers={"Referer": referer})
        if resp.status_code != requests.codes.OK:
            raise exceptions.PageLoadError(f"HTTP code {resp.status_code}")

        return resp.text

    def _get(self, *args, **kwargs):
        if "timeout" not in kwargs:
            kwargs["timeout"] = self.timeout

        try:
            return self.session.get(*args, **kwargs)
        except requests.RequestException as e:
            raise exceptions.NetworkError(e)

    REAUTH_SUBSTR = 'cookie="Authorization=;path=/"'

    @classmethod
    def _is_reauth_doc(cls, doc) -> bool:
        first_script = find_scripts(doc)[0]
        return cls.REAUTH_SUBSTR in first_script



from dataclasses import dataclass

from tplink_wr.parse.utils import extract_vars
from tplink_wr.router import RouterInterface

from .fetcher import Fetcher


@dataclass
class WLANStats(Fetcher):
    ssid: list[str]
    mac_filter_enabled: bool
    mac_filter_whitelist: bool
    clients: list

    @classmethod
    def fetch(cls, router: RouterInterface):
        last = stats_raw = cls._load_page(router, 1)

        while not last["last_page"]:
            last = cls._load_page(router, last["page_num"]+1)
            stats_raw["clients"] += last["clients"]

        stats = cls(
            ssid=[
                str(ssid)
                for ssid in stats_raw["ssid"]
            ],
            mac_filter_enabled=bool(
                stats_raw["mac_filter_enabled"]
            ),
            mac_filter_whitelist=bool(
                stats_raw["mac_filter_whitelist"]
            ),
            clients=stats_raw["clients"],
        )
        return stats

    @staticmethod
    def _load_page(router: RouterInterface, page: int) -> dict:
        doc = router.page("WlanStationRpm", params={"Page": page})
        wlan_para, host_list, ssid_list = extract_vars(doc, [
            "wlanHostPara", "hostList", "ssidList"
        ]).values()

        clients_count = wlan_para[0]
        limit_per_page = wlan_para[2]
        params_per_client = wlan_para[4]

        stats = {
            "page_num": page,
            "last_page": False,

            "ssid": ssid_list,
            "mac_filter_enabled": wlan_para[5],
            "mac_filter_whitelist": wlan_para[6],
            "clients_count": clients_count,
            "clients": [],
        }

        clients_left = clients_count - (page - 1) * limit_per_page
        this_page_count = min(clients_left, limit_per_page)
        for i in range(this_page_count):
            base = i * params_per_client
            stats["clients"].append({
                "mac": host_list[base],
                "rx": host_list[base + 2],
                "tx": host_list[base + 3],
            })

        if clients_left <= limit_per_page:
            stats["last_page"] = True

        return stats



from abc import ABC, abstractmethod
from dataclasses import asdict

from tplink_wr.router import RouterInterface


class Fetcher(ABC):
    @staticmethod
    @abstractmethod
    def fetch(router: RouterInterface):
        pass

    def dict(self):
        return asdict(self)



class CommunicationError(Exception):
    pass


class NetworkError(CommunicationError):
    pass


class ResponseError(CommunicationError):
    pass


class AuthError(ResponseError):
    pass


class PageLoadError(ResponseError):
    pass



#!/usr/bin/env python
from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="tplink-wr-api",
    version="0.2.1",
    url="https://github.com/n1k0r/tplink-wr-api",
    author="n1k0r",
    author_email="me@n1k0r.me",
    description="API to some budget TP-Link routers",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Telecommunications Industry",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Communications",
        "Topic :: Software Development :: Libraries",
        "Topic :: System :: Networking",
    ],

    packages=find_packages(exclude=["tests", "tests.*"]),
    python_requires=">=3.8",
    install_requires=[
        "requests~=2.26",
    ],
)



from html.parser import HTMLParser


class ScriptFinder(HTMLParser):
    def __init__(self):
        super().__init__()

        self.scripts = []
        self.script_in_progress = False

    def get_scripts(self):
        return self.scripts

    def close(self):
        super().close()

        self.scripts.clear()
        self.script_in_progress = False

    def handle_starttag(self, tag, attrs):
        if tag != "script":
            return

        self.script_in_progress = True
        self.current_script = ""

    def handle_endtag(self, tag):
        if tag != "script":
            return

        self.script_in_progress = False
        if self.current_script.strip():
            self.scripts.append(self.current_script)
        self.current_script = ""

    def handle_data(self, data):
        if not self.script_in_progress:
            return

        self.current_script += data


def find_scripts(data):
    finder = ScriptFinder()
    finder.feed(data)
    scripts = finder.get_scripts()
    return scripts



import json
import re


def find_var(name: str, script: str):
    match = re.search(f"\s+{name}\s*=\s*([^\s]([\s\S]*?[^\s])??)\s*;", script)
    if not match:
        return None
    return match.group(1)


def parse_var(var: str, **kwargs):
    if (array := parse_array(var, **kwargs)) is not None:
        return array

    return parse_primitive(var, **kwargs)


def parse_primitive(value: str, **kwargs):
    try:
        result = json.loads(value)
    except json.decoder.JSONDecodeError:
        return None

    return result


def parse_array(value: str, strip_zeros=True, **kwargs):
    if not (match := re.match(f"new\s+Array\(([\s\S]*)\)", value)):
        return None

    content = match.group(1)
    value_wrapped = f"[{content}]"
    values = parse_primitive(value_wrapped)

    if strip_zeros and list(values[-2:]) == [0, 0]:
        values = values[:-2]

    return values


def get_var(name: str, script: str, **kwargs):
    var = find_var(name, script)
    if not var:
        return None

    value = parse_var(var, **kwargs)
    return value
